import asyncio
import sys

_QT_KICKER_INSTALLED = {}


def install_qt_kicker(loop=None, update_rate=0.03):
    """Install a periodic callback to integrate Qt and asyncio event loops.

    If a version of the Qt bindings are not already imported, this function
    will do nothing.

    It is safe to call this function multiple times.

    Parameters
    ----------
    loop : event loop, optional
    update_rate : number
        Seconds between periodic updates. Default is 0.03.
    """
    if loop is None:
        loop = asyncio.get_event_loop()
    global _QT_KICKER_INSTALLED
    if loop in _QT_KICKER_INSTALLED:
        return
    if not any(p in sys.modules for p in ['PyQt4', 'pyside', 'PyQt5']):
        return
    import matplotlib.backends.backend_qt5
    from matplotlib.backends.backend_qt5 import _create_qApp
    from matplotlib._pylab_helpers import Gcf

    _create_qApp()
    qApp = matplotlib.backends.backend_qt5.qApp

    try:
        _draw_all = Gcf.draw_all  # mpl version >= 1.5
    except AttributeError:
        # slower, but backward-compatible
        def _draw_all():
            for f_mgr in Gcf.get_all_fig_managers():
                f_mgr.canvas.draw_idle()

    def _qt_kicker():
        # The RunEngine Event Loop interferes with the qt event loop. Here we
        # kick it to keep it going.
        _draw_all()

        qApp.processEvents()
        loop.call_later(update_rate, _qt_kicker)

    _QT_KICKER_INSTALLED[loop] = loop.call_soon(_qt_kicker)



from pkg_resources import resource_filename

import pdfstream.callbacks.calibration as mod

fn = resource_filename("tests", "configs/xpd_server.ini")


def test_Calibration(db_with_dark_and_calib, tmpdir):
    db = db_with_dark_and_calib
    config = mod.CalibrationConfig()
    config.read(fn)
    config.tiff_base = str(tmpdir)
    config.calib_base = str(tmpdir)
    cb = mod.Calibration(config, test=True)
    for name, doc in db[-1].canonical(fill="yes"):
        cb(name, doc)
    assert len(list(config.tiff_base.joinpath("calib").rglob("*.tiff"))) > 0


def test_Calibration_error(db_with_dark_and_calib, tmpdir):
    """Test error message when wavelength is None."""
    db = db_with_dark_and_calib
    config = mod.CalibrationConfig()
    config.read(fn)
    config.tiff_base = str(tmpdir)
    config.calib_base = str(tmpdir)
    cb = mod.Calibration(config, test=True)
    for name, doc in db[-1].canonical(fill="yes"):
        if name == "start":
            doc = dict(**doc)
            doc.update({"bt_wavelength": None})
        cb(name, doc)



import sys
from os import path

from setuptools import setup, find_packages

# NOTE: This file must remain Python 2 compatible for the foreseeable future,
# to ensure that we error out properly for people with outdated setuptools
# and/or pip.
min_version = (3, 7)
if sys.version_info < min_version:
    error = ("\n"
             "pdfstream does not support Python {0}.{1}.\n"
             "Python {2}.{3} and above is required. Check your Python version like so:\n"
             "\n"
             "python3 --version\n"
             "\n"
             "This may be due to an out-of-date pip. Make sure you have pip >= 9.0.1.\n"
             "Upgrade pip like so:\n"
             "\n"
             "pip install --upgrade pip\n").format(*(sys.version_info[:2] + min_version))
    sys.exit(error)

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as readme_file:
    readme = readme_file.read()

setup(
    name='pdfstream',
    version='0.4.3',
    description="The configs analysis toolbox for the study on pair distribution function (PDF).",
    install_requires=[],
    long_description=readme,
    long_description_content_type='text/markdown',
    author="Songsheng Tao",
    author_email='st3107@columbia.edu',
    url='https://github.com/xpdAcq/pdfstream',
    python_requires='>={}'.format('.'.join(str(n) for n in min_version)),
    packages=find_packages(exclude=['docs', 'tests']),
    entry_points={
        'console_scripts': [
            'pdfstream = pdfstream.main:main',
            'run_server = pdfstream.main:run_server',
            'print_server_config_dir = pdfstream.main:print_server_config_dir'
            # 'command = some.module:some_function',
        ],
    },
    include_package_data=True,
    package_data={
        'pdfstream': [
            'configs/*'
            # When adding files here, remember to update MANIFEST.in as well,
            # or else they will not be included in the distribution on PyPI!
            # 'path/to/data_file',
        ]
    },
    license="BSD (3-clause)",
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
    ],
    scripts=[
        'scripts/pdfstream_install'
    ]
)



import itertools
import typing as tp

import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from numpy import ndarray

from pdfstream.visualization.tools import plot_line, plot_fit, normalize, shift, auto_text, set_minor_tick, \
    auto_label

PLOT_METHOD = {
    "line": plot_line,
    "fit": plot_fit
}
TEXT_IND = {
    "line": -1,
    "fit": -4
}


def _waterfall(
    data: ndarray, plot_method: tp.Callable, ax: Axes, normal: bool = True,
    stack: bool = True, gap: float = 0, text: tp.Union[str, None] = None, text_xy: tuple = None,
    text_ind: int = -1, **kwargs
) -> None:
    """The basic visualization function to realize single, waterfall, and comparison plot.

    Parameters
    ----------
    data : ndarray
        The data array. The first row is the independent variable and the following rows are the dependent
        variable.

    plot_method : Callable
        A callable that has the usage like 'plot_method(data: ndarray, ax: Axes, **kwargs)'

    ax : Axes
        The axes to visualize the data. If None, use current axes.

    normal : bool
        If True, the second and the following rows in data will be normalized by (max - min). Else, do nothing.

    stack : bool
        If True, the second and the third rows will be shifted so that there will be a gap between data.

    gap : float
        The gap between the adjacent curves. It is defined by the nearest points in vertical direction.

    text : str
        The text to annotate the curve.

    text_xy : tuple
        The tuple of x and y position of the annotation in data coordinates. If None, use the default in the
        'tools.auto_text'.

    text_ind : int
        The index of the lines in the axes to be annotated after the data is plotted.

    kwargs : optional
        The kwargs arguments for the 'plot_method'.
    """
    if normal:
        data = normalize(data)
    if stack:
        data = shift(data, ax, gap=gap)
    plot_method(data, ax, **kwargs)
    if text is not None:
        auto_text(text, ax, text_xy=text_xy, index=text_ind)
    return


def waterfall(
    dataset: tp.Iterable[ndarray], ax: Axes = None, mode: str = "line", normal: bool = True,
    stack: bool = True, gap: float = 0, texts: tp.Iterable[str] = None, text_xy: tuple = None,
    label: str = None, minor_tick: tp.Union[int, None] = 2, legends: tp.List[str] = None,
    colors: tp.Iterable = None, **kwargs
) -> Axes:
    """The visualization function to realize waterfall, and comparison plot.

    Parameters
    ----------
    dataset :
        A iterable of the data array. The requirement for dimensions depends on the mode.
        If mode = 'line', data = (x_array, y_array)
        If mode = 'fit', data = (x_array, y_array, ycalc_array)

    kwargs :
        The kwargs arguments for the plotting of each data. It depends on mode.
        If mode = 'line', kwargs in ('xy_kwargs',).
        If mode = 'fit', kwargs in ('xy_kwargs', 'xycalc_kwargs', 'xydiff_kwargs', 'xyzero_kwargs',
        'fill_kwargs', 'yzero').

    mode :
        The plotting mode. Currently support 'line', 'fit'.

    ax :
        The axes to visualize the data. If None, use current axes.

    normal :
        If True, the second and the following rows in data will be normalized by (max - min). Else, do nothing.

    stack :
        If True, the second and the third rows will be shifted so that there will be a gap between data (
        waterfall plot). Else, the data will be plotted without shifting (comparison plot).

    gap :
        The gap between the adjacent curves. It is defined by the nearest points in vertical direction.

    texts :
        The texts to annotate the curves. It has the same order as the curves.

    text_xy :
        The tuple of x and y position of the annotation in data coordinates. If None, use the default in the
        'tools.auto_text'.

    label :
        The label type used in automatic labeling. Acceptable types are listed in 'tools._LABELS'

    minor_tick :
        How many parts that the minor ticks separate the space between the two adjacent major ticks. Default 2.
        If None, no minor ticks.

    legends :
        The legend labels for the curves.

    colors :
        The color of the plots. It will be the value for the key 'color' in 'xy_kwargs' in kwargs. If None,
        use default color.

    kwargs :
        The key words for the 'plot_method'. The

    Returns
    -------
    ax : Axes
        The axes with the plot inside.
    """

    def _inject_color(_kwargs, _color):
        if 'xy_kwargs' in _kwargs:
            _kwargs['xy_kwargs'].update({'color': _color})
        else:
            _kwargs['xy_kwargs'] = {'color': _color}
        return

    if ax is None:
        ax = plt.gca()
    if texts is None:
        texts = tuple()
    if colors is None:
        colors = tuple()
    if mode not in PLOT_METHOD:
        raise ValueError(
            "Unknown mode {}. Mode options: {}".format(mode, list(PLOT_METHOD.keys()))
        )
    text_ind = TEXT_IND.get(mode, -1)
    for data, text, color in itertools.zip_longest(dataset, texts, colors):
        if color:
            _inject_color(kwargs, color)
        _waterfall(
            data, PLOT_METHOD[mode], ax=ax, normal=normal, stack=stack, gap=gap, text=text, text_xy=text_xy,
            text_ind=text_ind, **kwargs
        )
    if minor_tick is not None:
        set_minor_tick(ax, minor_tick)
    if label is not None:
        auto_label(ax, label)
    if legends is not None:
        ax.legend(legends)
    return ax


def visualize(
    data: ndarray, ax: Axes = None, mode: str = "line", normal: bool = False,
    text: str = None, text_xy: tuple = None, label: str = None,
    minor_tick: int = 2, legends: tp.List[str] = None, color: tp.Any = None, **kwargs
) -> Axes:
    """The visualization function to realize single plot.

    Parameters
    ----------
    data :
        A data array. The requirement for dimensions depends on the mode.
        If mode = 'line', data = (x_array, y_array)
        If mode = 'fit', data = (x_array, y_array, ycalc_array)

    kwargs :
        The kwargs arguments for the plotting of each data. It depends on mode.
        If mode = 'line', kwargs in ('xy_kwargs',).
        If mode = 'fit', kwargs in ('xy_kwargs', 'xycalc_kwargs', 'xydiff_kwargs', 'xyzero_kwargs',
        'fill_kwargs', 'yzero').

    mode :
        The plotting mode. Currently support 'line', 'fit'.

    ax :
        The axes to visualize the data. If None, use current axes.

    normal :
        If True, the second and the following rows in data will be normalized by (max - min). Else, do nothing.

    text :
        The text to annotate the curve.

    text_xy :
        The tuple of x and y position of the annotation in data coordinates. If None, use the default in the
        'tools.auto_text'.

    label :
        The label type used in automatic labeling. Acceptable types are listed in 'tools._LABELS'

    minor_tick : int
        How many parts that the minor ticks separate the space between the two adjacent major ticks. Default 2.
        If None, no minor ticks.

    legends :
        The legend label for the curve.

    color :
        The color of the plots. If None, use default color cycle in rc.

    Returns
    -------
    ax :
        The axes with the plot inside.
    """

    return waterfall(
        (data,), ax=ax, mode=mode, normal=normal, texts=(text,), text_xy=text_xy,
        label=label, minor_tick=minor_tick, stack=False,
        legends=legends if legends else None,
        colors=(color,), **kwargs
    )



"""The analysis server. Process raw image to PDF."""
import typing as tp

import databroker
from bluesky.callbacks.zmq import Publisher
from databroker.v2 import Broker
from event_model import RunRouter
from ophyd.sim import NumpySeqHandler

import pdfstream.io as io
from pdfstream.callbacks.analysis import AnalysisConfig, VisConfig, ExportConfig, AnalysisStream, Exporter, \
    Visualizer, no_need_to_refresh_db, ExporterXpdan
from pdfstream.callbacks.calibration import CalibrationConfig, Calibration
from pdfstream.servers import CONFIG_DIR, ServerNames
from pdfstream.servers.base import ServerConfig, find_cfg_file, BaseServer, StartStopCallback


class XPDConfig(AnalysisConfig, VisConfig, ExportConfig, CalibrationConfig):
    """The configuration for the xpd data reduction. It consists of analysis, visualization and exportation."""

    def __init__(self, *args, **kwargs):
        super(XPDConfig, self).__init__(*args, **kwargs)
        self._an_db = None

    @property
    def an_db(self) -> tp.Union[None, Broker]:
        name = self.get("DATABASE", "an_db", fallback=None)
        if no_need_to_refresh_db(self._an_db, name):
            pass
        elif name is None:
            self._an_db = None
        elif name == "temp":
            print("Warning: a temporary db is created for an db. It will be destroy at the end of the session.")
            self._an_db = databroker.v2.temp()
        else:
            self._an_db = databroker.catalog[name]
        return self._an_db

    @an_db.setter
    def an_db(self, db: Broker):
        section_name = "DATABASE"
        db_key = "an_db"
        if section_name not in self.sections():
            self.add_section(section_name)
        self.set(section_name, db_key, db.name)
        self._an_db = db

    @property
    def publisher_config(self) -> dict:
        host = self.get("PUBLISH TO", "host")
        port = self.getint("PUBLISH TO", "port")
        prefix = self.get("PUBLISH TO", "prefix", fallback="").encode()
        return {
            "address": (host, port),
            "prefix": prefix
        }

    @property
    def functionality(self) -> dict:
        return {
            "do_calibration": self.getboolean("FUNCTIONALITY", "do_calibration"),
            "dump_to_db": self.getboolean("FUNCTIONALITY", "dump_to_db"),
            "export_files": self.getboolean("FUNCTIONALITY", "export_files"),
            "visualize_data": self.getboolean("FUNCTIONALITY", "visualize_data"),
            "send_messages": self.getboolean("FUNCTIONALITY", "send_messages"),
            "export_files_in_xpdan_style": self.getboolean("FUNCTIONALITY", "export_files_in_xpdan_style")
        }


class XPDServerConfig(XPDConfig, ServerConfig):
    """The configuration for xpd server."""
    pass


class XPDServer(BaseServer):
    """The server of XPD data analysis. It is a live dispatcher with XPDRouter subscribed."""
    def __init__(self, config: XPDServerConfig):
        super(XPDServer, self).__init__(config.address, prefix=config.prefix)
        self.subscribe(StartStopCallback())
        self.subscribe(XPDRouter(config))


def make_and_run(
    cfg_file: str = None,
    *,
    suppress_warning: bool = True
):
    """Run the xpd data reduction server.

    The server will receive message from proxy and process the data in the message. The processed data will be
    visualized and exported to database and the file system.

    Parameters
    ----------
    cfg_file :
        The path to configuration .ini file. The default path is "~/.config/acq/xpd_server.ini".

    suppress_warning :
        If True, all warning will be suppressed. Turn it to False when running in a test.
    """
    if suppress_warning:
        import warnings
        warnings.simplefilter("ignore")
    if not cfg_file:
        cfg_file = find_cfg_file(CONFIG_DIR, ServerNames.xpd)
    config = XPDServerConfig(allow_no_value=True)
    config.read(cfg_file)
    server = XPDServer(config)
    server.install_qt_kicker()
    server.start()


class XPDRouter(RunRouter):
    """A router that contains the callbacks for the xpd data reduction."""

    def __init__(self, config: XPDConfig):
        factory = XPDFactory(config)
        super(XPDRouter, self).__init__(
            [factory],
            handler_registry={"NPY_SEQ": NumpySeqHandler}
        )


class XPDFactory:
    """The factory to generate callback for xpd data reduction."""

    def __init__(self, config: XPDConfig):
        self.config = config
        self.functionality = self.config.functionality
        self.analysis = [AnalysisStream(config)]
        self.calibration = [Calibration(config)] if self.functionality["do_calibration"] else []
        if self.functionality["dump_to_db"]:
            self.analysis[0].subscribe(self.config.an_db.v1.insert)
        if self.functionality["export_files"]:
            self.analysis[0].subscribe(Exporter(config))
        if self.functionality["visualize_data"]:
            self.analysis[0].subscribe(Visualizer(config))
        if self.functionality["send_messages"]:
            self.analysis[0].subscribe(Publisher(**self.config.publisher_config))
        if self.functionality["export_files_in_xpdan_style"]:
            self.analysis[0].subscribe(ExporterXpdan(config))

    def __call__(self, name: str, doc: dict) -> tp.Tuple[list, list]:
        if name == "start":
            if doc.get(self.config.dark_identifier):
                # dark frame run
                io.server_message("Ignore dark frame.")
                return [], []
            elif doc.get(self.config.calib_identifier):
                # calibration run
                io.server_message("Start calibration.")
                return self.calibration, []
            else:
                # light frame run
                io.server_message("Start data reduction.")
                return self.analysis, []
        return [], []



from pathlib import Path

from diffpy.pdfgetx import PDFConfig, PDFGetter


def load_pdfconfig(cfg_file: str) -> PDFConfig:
    """Load the PDFConfig from the processed data file or configuration file."""
    pdfconfig = PDFConfig()
    pdfconfig.readConfig(cfg_file)
    return pdfconfig


def write_pdfgetter(saving_dir: str, filename: str, pdfgetter: PDFGetter) -> dict:
    """Write out data in pdfgetter into files"""
    data_dirs = {}
    for out_type in pdfgetter.config.outputtypes:
        data_dir = Path(saving_dir).joinpath(out_type)
        if not data_dir.exists():
            data_dir.mkdir()
        data_dirs.update({out_type: data_dir})
    dct = {}
    for out_type in pdfgetter.config.outputtypes:  # out_type in ('iq', 'sq', 'fq', 'gr')
        data_dir = data_dirs.get(out_type)
        out_file = data_dir.joinpath(Path(filename).with_suffix(".{}".format(out_type)).name)
        pdfgetter.writeOutput(str(out_file), out_type)
        dct.update({out_type: str(out_file)})
    return dct



import typing
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed
from pathlib import Path

import matplotlib.pyplot as plt

from pdfstream.io import load_array
from .io import load_pdfconfig, write_pdfgetter
from .main import get_pdf


def transform(
    cfg_file: str,
    *data_files: str,
    output_dir: str = ".",
    plot_setting: typing.Union[str, dict] = None,
    parallel: bool = False,
    test: bool = False
) -> typing.List[typing.Dict[str, str]]:
    """Transform the data to I(Q), S(Q), F(Q), G(r).

    This is a simple interface for diffpy.pdfgetx. For more interactive usage, please use `pdfgetx3` command.

    Parameters
    ----------
    cfg_file :
        The configure file for diffpy.pdfgetx.

    data_files :
        The single or multiple data files.

    output_dir :
        The output directory for files. The file will be saved in the sub-directories.

    plot_setting :
        The dictionary for the visualization setting. The keys are for the plot in matplotlib
        (https://matplotlib.org/api/_as_gen/matplotlib.axes.Axes.plot.html).

    parallel :
        If True, use parallel computing.

    test :
        If True, use test mode (for developers).

    Returns
    -------
    dcts :
        A list of dictionaries. The keys are the output data types and the values are the output file paths.
    """
    Path(output_dir).mkdir(parents=True, exist_ok=True)
    if parallel:
        executor = ProcessPoolExecutor() if not test else ThreadPoolExecutor()
        jobs = [
            executor.submit(
                _transform, cfg_file, data_file, output_dir=output_dir, plot_setting=plot_setting, test=test
            )
            for data_file in data_files
        ]
        return [job.result() for job in as_completed(jobs)]
    return [
        _transform(
            cfg_file, data_file, output_dir=output_dir, plot_setting=plot_setting, test=test
        )
        for data_file in data_files
    ]


def _transform(
    cfg_file: str,
    data_file: str,
    output_dir: str = ".",
    plot_setting: typing.Union[str, dict] = None,
    test: bool = False,
) -> typing.Dict[str, str]:
    """Transform the data."""
    pdfconfig = load_pdfconfig(cfg_file)
    chi = load_array(data_file)
    pdfgetter = get_pdf(pdfconfig, chi, plot_setting=plot_setting)
    filename = Path(data_file).stem
    dct = write_pdfgetter(output_dir, filename, pdfgetter)
    if not test:
        plt.show()
    return dct



from configparser import Error
from pathlib import Path

import matplotlib.pyplot as plt
import pytest
from pkg_resources import resource_filename

import pdfstream.callbacks
import pdfstream.callbacks.analysis as an
from pdfstream.schemas import analysis_out_schemas, analysis_in_schemas, Validator

fn = resource_filename("tests", "configs/xpd_server.ini")


@pytest.mark.parametrize("use_db", [True, False])
def test_AnalysisStream(db_with_img_and_bg_img, use_db):
    db = db_with_img_and_bg_img
    config = an.AnalysisConfig()
    config.read(fn)
    config.raw_db = db
    ld = an.AnalysisStream(config)
    # validate that output data
    out_validator = Validator(analysis_out_schemas)
    ld.subscribe(out_validator)
    # validate the input data
    in_validator = Validator(analysis_in_schemas)
    for name, doc in db[-1].canonical(fill="yes", strict_order=True):
        in_validator(name, doc)
        ld(name, doc)


def test_Visualizer(db_with_dark_and_scan):
    db = db_with_dark_and_scan
    config = an.AnalysisConfig()
    config.raw_db = db
    config.read(fn)
    ld = an.AnalysisStream(config)
    config1 = pdfstream.callbacks.analysis.VisConfig()
    config1.read(fn)
    config1.fig = plt.figure()
    cb = pdfstream.callbacks.analysis.Visualizer(config1)
    ld.subscribe(cb)
    for name, doc in db[-1].canonical(fill="yes", strict_order=True):
        ld(name, doc)
    cb.show_figs()


def test_Exporter(db_with_dark_and_scan, tmpdir):
    db = db_with_dark_and_scan
    config = an.AnalysisConfig()
    config.read(fn)
    config.raw_db = db
    ld = an.AnalysisStream(config)
    ep_config = pdfstream.callbacks.analysis.ExportConfig()
    ep_config.read(fn)
    ep_config.tiff_base = str(tmpdir)
    ep = pdfstream.callbacks.analysis.Exporter(ep_config)
    ld.subscribe(ep)
    for name, doc in db[-1].canonical(fill="yes", strict_order=True):
        ld(name, doc)
    tiff_base = Path(ep_config.tiff_base)
    assert len(list(tiff_base.rglob("*.tiff"))) > 0
    assert len(list(tiff_base.rglob("*.csv"))) > 0
    assert len(list(tiff_base.rglob("*.json"))) > 0


def test_ExporterXpdan(db_with_dark_and_scan, tmpdir):
    """Test ExporterXpaan. It should output the correct files in a two layer directory."""
    db = db_with_dark_and_scan
    config = an.AnalysisConfig()
    config.read(fn)
    config.raw_db = db
    ld = an.AnalysisStream(config)
    ep_config = pdfstream.callbacks.analysis.ExportConfig()
    ep_config.read(fn)
    ep_config.tiff_base = str(tmpdir)
    ep = pdfstream.callbacks.analysis.ExporterXpdan(ep_config)
    ld.subscribe(ep)
    for name, doc in db[-1].canonical(fill="yes", strict_order=True):
        ld(name, doc)
    tiff_base = Path(ep_config.tiff_base)
    data_folder = tiff_base.joinpath("Ni")
    assert data_folder.is_dir()
    for dir_name in ("dark_sub", "integration", "meta", "mask", "iq", "sq", "fq", "pdf", "scalar_data"):
        assert data_folder.joinpath(dir_name).is_dir()
        assert len(list(data_folder.joinpath(dir_name).glob("*.*"))) > 0


def test_ExportConfig():
    config = pdfstream.callbacks.analysis.ExportConfig()
    config.read(fn)
    with pytest.raises(Error):
        assert config.tiff_base



"""The tools to read, write, audit, transform, wash the data."""
from typing import Any
from typing import Callable
from typing import Generator


def paths(dct: Any, path=()) -> Generator:
    """Yield paths to the leafs of a nested dictionary."""
    if isinstance(dct, dict):
        for key, value in dct.items():
            yield from paths(value, path + (key,))
    elif isinstance(dct, (tuple, list)):
        for ind, elem in enumerate(dct):
            yield from paths(elem, path + (ind,))
    else:
        yield path + (dct,)


def iter_dct(dct: dict, operation: Callable):
    """Recursively iterate the dictionary and operate on the leaf nodes. Return the processed dictionary."""
    dct2 = dict()
    for key, value in dct.items():
        if isinstance(value, dict):
            dct2[key] = iter_dct(value, operation)
        else:
            dct2[key] = operation(value)
    return dct2


def to_dict(obj: object, exclude: Callable[[str], bool] = None) -> dict:
    """Convert an object to a dictionary by keeping import attribute information."""

    def pdfconfig_exclude(attr: str) -> bool:
        """Whether to exclude an attribute in pdfconfig."""
        if attr.startswith("_"):
            return True
        if attr in ("inputfiles", "output", "interact"):
            return True
        return False

    dct = {}
    if exclude is None:
        exclude = pdfconfig_exclude
    for attr_key, attr_val in obj.__dict__.items():
        if not exclude(attr_key):
            dct[attr_key] = attr_val
    return dct


def get_value(dct: dict, keys: tuple) -> Any:
    """Get the value by walking along a key chain."""
    result = dct
    for key in keys:
        result = result[key]
    return result



import typing as tp

from databroker import catalog
from databroker.core import BlueskyRun

from pdfstream.analyzers.base import AnalyzerConfig, Analyzer
from pdfstream.servers.xpd_server import XPDRouter, XPDConfig


class XPDAnalyzerConfig(XPDConfig, AnalyzerConfig):
    """The configuration of the XPDAnalyzer."""
    pass


class XPDAnalyzer(XPDRouter, Analyzer):
    """The class to process the XPD data to PDF data. """
    pass


def replay(run: BlueskyRun) -> tp.Tuple[XPDAnalyzerConfig, XPDAnalyzer]:
    """Generate the original data, original configure and the XPD analyzer of it.

    Parameters
    ----------
    run :
        The run containing the processed data.

    Returns
    -------
    config :
        The original configuration.

    analyzer :
        The original analyzer.
    """
    config = XPDAnalyzerConfig()
    config.read_run(run)
    analyzer = XPDAnalyzer(config)
    return config, analyzer


def retrieve_original_run(run: BlueskyRun) -> tp.Union[None, BlueskyRun]:
    """Retrieve the original run."""
    start = run.metadata['start']
    if 'original_run_uid' not in start:
        raise Warning("Missing original_run_uid. Cannot retrieve original run.")
    if 'original_db' not in start:
        raise Warning("Missing original_db. Cannot retrieve original run.")
    try:
        db = catalog[start['original_db']]
    except KeyError:
        raise Warning("Missing {} in catalog. Cannot retrieve original run.".format(start['original_db']))
    try:
        return db[start['original_run_uid']]
    except KeyError:
        raise Warning("Run {} not found in database.".format(start['original_run_uid']))
